import importlib
import traceback
import logging
import json
from typing import Dict, Any, List
import asyncio
from datetime import datetime, timedelta
import re
import os
from .actfile_parser import ActfileParser, ActfileParserError

# Configure logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

class ExecutionManager:
    def __init__(self, actfile_path: str = 'Actfile'):
        logger.info(f"Initializing ExecutionManager")
        self.actfile_path = actfile_path
        self.node_results = {}
        self.execution_queue = asyncio.Queue()
        self.load_workflow()

    def load_workflow(self):
        logger.info("Loading workflow data")
        try:
            parser = ActfileParser(self.actfile_path)
            self.workflow_data = parser.parse()
            self.actfile_parser = parser
        except ActfileParserError as e:
            logger.error(f"Error parsing Actfile: {e}")
            raise

        self.load_node_executors()

    def load_node_executors(self):
        logger.info("Loading node executors")
        node_types = set(node['type'] for node in self.workflow_data['nodes'].values())
        self.node_executors = {}
        for node_type in node_types:
            try:
                logger.info(f"Attempting to load node type: {node_type}")
                module_name = f"act.nodes.{node_type.lower()}_node"
                module = importlib.import_module(module_name)
                logger.info(f"Module loaded: {module}")

                class_names = [
                    f"{node_type}Node",
                    f"{node_type.capitalize()}Node",
                    f"{node_type}NodeNode",
                    node_type
                ]

                for class_name in class_names:
                    logger.info(f"Looking for class: {class_name}")
                    if hasattr(module, class_name):
                        node_class = getattr(module, class_name)
                        logger.info(f"Class found: {node_class}")
                        node_instance = node_class()
                        if hasattr(node_instance, 'set_execution_manager'):
                            node_instance.set_execution_manager(self)
                        self.node_executors[node_type] = node_instance
                        logger.info(f"Executor created for {node_type}")
                        break
                else:
                    logger.warning(f"No suitable class found for {node_type}")
                    raise AttributeError(f"No suitable class found for {node_type}")

            except Exception as e:
                logger.error(f"Error loading node type '{node_type}': {str(e)}")
                logger.error(traceback.format_exc())
                from act.nodes.generic_node import GenericNode
                self.node_executors[node_type] = GenericNode()

    def execute_workflow(self) -> Dict[str, Any]:
        logger.info(f"Starting execution of workflow")
        self.node_results = {}
        execution_queue = []

        try:
            start_node_name = self.actfile_parser.get_start_node()
            if not start_node_name:
                logger.error("No start node specified in Actfile.")
                return {"status": "error", "message": "No start node specified in Actfile.", "results": {}}

            execution_queue.append((start_node_name, None))

            while execution_queue:
                node_name, input_data = execution_queue.pop(0)
                node_result = self.execute_node(node_name, input_data)
                self.node_results[node_name] = node_result

                successors = self.actfile_parser.get_node_successors(node_name)
                for successor in successors:
                    logger.debug(f"Queueing next node: {successor}")
                    execution_queue.append((successor, node_result))

            logger.info("Workflow execution completed")
            return {
                "status": "success",
                "message": "Workflow executed successfully",
                "results": self.node_results
            }

        except Exception as e:
            logger.error(f"Error during workflow execution: {str(e)}", exc_info=True)
            return {
                "status": "error",
                "message": f"Workflow execution failed: {str(e)}",
                "results": self.node_results
            }

    def execute_node(self, node_name: str, input_data: Dict[str, Any] = None):
        logger.info(f"Executing node: {node_name}")
        try:
            node = self.workflow_data['nodes'][node_name]
            node_type = node.get('type')
            node_data = node.copy()

            if input_data:
                node_data['input'] = input_data

            resolved_node_data = self.resolve_placeholders_for_execution(node_data)

            logger.info(f"Node type: {node_type}")
            logger.info(f"Node data after resolving placeholders: {self.log_safe_node_data(resolved_node_data)}")

            executor = self.node_executors.get(node_type)
            if executor:
                logger.info(f"Executor found for node type: {node_type}")

                result = executor.execute(resolved_node_data)
                logger.info(f"Node {node_name} execution result: {self.log_safe_node_data(result)}")

                self.node_results[node_name] = result
                return result
            else:
                logger.error(f"No executor found for node type: {node_type}")
                result = {"status": "error", "message": f"No executor found for node type: {node_type}"}
                self.node_results[node_name] = result
                return result

        except Exception as e:
            logger.error(f"Error executing node {node_name}: {str(e)}")
            logger.error(traceback.format_exc())
            error_result = {"status": "error", "message": str(e)}
            self.node_results[node_name] = error_result
            return error_result

    def resolve_placeholders_for_execution(self, data: Any) -> Any:
        if isinstance(data, dict):
            return {k: self.resolve_placeholders_for_execution(v) for k, v in data.items()}
        elif isinstance(data, list):
            return [self.resolve_placeholders_for_execution(item) for item in data]
        elif isinstance(data, str):
            return self.resolve_placeholder_string(data)
        return data

    def resolve_placeholder_string(self, text: str) -> str:
        # Handle environment variables
        if text.startswith('${') and text.endswith('}'):
            env_var = text[2:-1]
            return os.environ.get(env_var, text)
        
        pattern = re.compile(r'\{\{(.*?)\}\}')
        matches = pattern.findall(text)
        
        for match in matches:
            parts = match.split('.')
            node_id = parts[0]
            path = '.'.join(parts[1:])
            value = self.fetch_value(node_id, path)
            if value is not None:
                text = text.replace(f"{{{{{match}}}}}", str(value))
        
        return text

    def fetch_value(self, node_id: str, path: str) -> Any:
        logger.info(f"Fetching value for node_id: {node_id}, path: {path}")
        if node_id in self.node_results:
            result = self.node_results[node_id]
            for part in path.split('.'):
                if isinstance(result, dict) and part in result:
                    result = result[part]
                else:
                    return None
            return result
        return None

    @staticmethod
    def log_safe_node_data(node_data):
        if isinstance(node_data, dict):
            safe_data = {k: ('[REDACTED]' if k == 'api_key' else v) for k, v in node_data.items()}
        else:
            safe_data = node_data
        return json.dumps(safe_data, indent=2)

    @classmethod
    def register_node_type(cls, node_type: str, node_class: Any):
        logger.info(f"Registering custom node type: {node_type}")
        if not hasattr(cls, 'custom_node_types'):
            cls.custom_node_types = {}
        cls.custom_node_types[node_type] = node_class

    def get_node_executor(self, node_type: str) -> Any:
        if hasattr(self, 'custom_node_types') and node_type in self.custom_node_types:
            return self.custom_node_types[node_type]()
        return self.node_executors.get(node_type)


import os
import json
import logging
from typing import Dict, Any, Union
import requests
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

class RequestNode:
    def __init__(self):
        logger.info("Initializing RequestNode")
        self.session = requests.Session()
        retry = Retry(total=3, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
        self.session.mount('http://', HTTPAdapter(max_retries=retry))
        self.session.mount('https://', HTTPAdapter(max_retries=retry))

    def execute(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        logger.info("Starting execution of RequestNode")
        logger.debug(f"Received node_data: {json.dumps(self.log_safe_node_data(node_data), indent=2)}")

        url = node_data.get('url', '').strip()
        method = node_data.get('method', 'GET').upper()
        headers = {'Content-Type': 'application/json'}
        auth_type = node_data.get('auth_type', 'none')
        auth_value = node_data.get('auth_value', '')
        parameters = node_data.get('parameters', {})
        body = node_data.get('body', {})
        timeout = node_data.get('timeout', 30)
        verify_ssl = node_data.get('verify_ssl', True)
        max_response_size = node_data.get('max_response_size', 10 * 1024 * 1024)  # 10 MB default

        if not url:
            logger.error("URL not provided in node data")
            return {"status": "error", "message": "URL not provided", "output": None}

        try:
            # Parse body if it's a string
            if isinstance(body, str):
                try:
                    body = json.loads(body)
                except json.JSONDecodeError:
                    logger.error("Failed to parse body as JSON")
                    return {"status": "error", "message": "Invalid JSON in body", "output": None}

            # Set authentication header
            if auth_type == 'api_key':
                headers['Authorization'] = auth_value

            # Make the request
            result = self.make_request(url, method, headers, parameters, body, timeout, verify_ssl, max_response_size)
            logger.info("Request completed successfully")
            return {
                "status": "success",
                "output": result
            }
        except Exception as e:
            logger.error(f"Error in making request: {str(e)}")
            return {"status": "error", "message": str(e), "output": None}

    def make_request(self, url: str, method: str, headers: Dict[str, str], 
                     parameters: Dict[str, Any], body: Dict[str, Any], 
                     timeout: int, verify_ssl: bool, max_response_size: int) -> Dict[str, Any]:
        logger.info(f"Making {method} request to {url}")
        
        try:
            # Prepare the request
            request_kwargs = {
                "method": method,
                "url": url,
                "headers": headers,
                "params": parameters,
                "timeout": timeout,
                "verify": verify_ssl
            }

            # Add body for POST, PUT, PATCH
            if method in ['POST', 'PUT', 'PATCH']:
                request_kwargs["json"] = body

            # Make the request
            with self.session.request(**request_kwargs) as response:
                response.raise_for_status()

                # Check response size
                if int(response.headers.get('Content-Length', 0)) > max_response_size:
                    raise ValueError(f"Response size exceeds maximum allowed size of {max_response_size} bytes")

                try:
                    result = response.json()
                except json.JSONDecodeError:
                    result = response.text

                return {
                    "status_code": response.status_code,
                    "headers": dict(response.headers),
                    "body": result,
                    "url": response.url
                }

        except requests.exceptions.RequestException as e:
            logger.error(f"Request failed: {str(e)}")
            raise

    @staticmethod
    def log_safe_node_data(node_data):
        if isinstance(node_data, dict):
            safe_data = {k: ('[REDACTED]' if k in ['auth_value', 'api_key'] else v) for k, v in node_data.items()}
        else:
            safe_data = node_data
        return safe_data

# This part is for testing the RequestNode independently
if __name__ == "__main__":
    # Make the specific API request
    node_data = {
        "url": "https://dobee-dev.ew.r.appspot.com/objectives/mi331cCJ0abn19PD1HkX/keyResults",
        "method": "GET",
        "auth_type": "api_key",
        "auth_value": "Bearer eyJhbGciOiJSUzI1NiIsImtpZCI6IjcxOGY0ZGY5MmFkMTc1ZjZhMDMwN2FiNjVkOGY2N2YwNTRmYTFlNWYiLCJ0eXAiOiJKV1QifQ.eyJpc3MiOiJodHRwczovL3NlY3VyZXRva2VuLmdvb2dsZS5jb20vZG9iZWUtZGV2IiwiYXVkIjoiZG9iZWUtZGV2IiwiYXV0aF90aW1lIjoxNzI4NTg5MjczLCJ1c2VyX2lkIjoiblFzWTNTQUhVbFhsOTcwSFk1amh2eER4Rk10MiIsInN1YiI6Im5Rc1kzU0FIVWxYbDk3MEhZNWpodnhEeEZNdDIiLCJpYXQiOjE3MjkxOTI0NzYsImV4cCI6MTcyOTE5NjA3NiwiZW1haWwiOiJhZG1pbkBleGFtcGxlLmNvbSIsImVtYWlsX3ZlcmlmaWVkIjpmYWxzZSwiZmlyZWJhc2UiOnsiaWRlbnRpdGllcyI6eyJlbWFpbCI6WyJhZG1pbkBleGFtcGxlLmNvbSJdfSwic2lnbl9pbl9wcm92aWRlciI6InBhc3N3b3JkIn19.sS0PWtI4q2eAogmI3Uqw3Vmh9i9VCPIHNa1x3eDcs8hQuICM4CSj9sUvQLdVdHhx6gLTXB6XzU8u58CfCGjz7SD7QAnqmVyle4eaHu7MuPnY__4ieAwmCnFutpaxuFXWhbeq84nUEXuvgZGJIr4g_nCDPqWWDfXg1dYn0n5KIlvgCWe3eGd0nuHUnoWWF6dFSgf8eYmoEdyJ3VvOP4975imL0XKBdtJjbEgo7N8sUA3r_wWxV9U-hL7a-i50vhMCZ49vT7ncPnRNUuZAhq_PL0T_N2qu5MAPbWXtHiPWsI6CM2UDikpSqZIvGIWNyjim5GhWOTQy1vuNfpa99L6oaw"
    }
    
    request_node = RequestNode()
    result = request_node.execute(node_data)
    print("API Request Result:", json.dumps(RequestNode.log_safe_node_data(result), indent=2))


import logging
from typing import Dict, Any
import os

logger = logging.getLogger(__name__)

class AINode:
    def __init__(self):
        logger.info("Initializing AINode")
        # Initialize any AI-related configurations or models here

    def execute(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        operation = node_data.get('operation')
        if operation == 'decision':
            return self.make_decision(node_data)
        elif operation == 'end':
            return self.end_workflow(node_data)
        else:
            raise ValueError(f"Unsupported operation: {operation}")

   
    def end_workflow(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        logger.info("Workflow has ended.")
        return {"status": "success", "message": "Workflow completed successfully.", "output": None}



import logging
from typing import Dict, Any
import openai
import os

logger = logging.getLogger(__name__)

class OpenAINode:
    def __init__(self):
        logger.info("Initializing OpenAINode")
        openai_api_key = os.environ.get('OPENAI_API_KEY')
        if not openai_api_key:
            raise ValueError("OPENAI_API_KEY environment variable not set")
        openai.api_key = openai_api_key

    def execute(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        operation = node_data.get('operation')
        if operation == 'parseBugReport':
            return self.parse_bug_report(node_data)
        elif operation == 'validateBugReport':
            return self.validate_bug_report(node_data)
        else:
            raise ValueError(f"Unsupported operation: {operation}")

    def parse_bug_report(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        model = node_data.get('model', 'gpt-3.5-turbo')
        prompt = node_data.get('prompt', 'Parse the following bug report:')
        bug_report = node_data.get('bug_report', '')
        try:
            response = openai.ChatCompletion.create(
                model=model,
                messages=[
                    {"role": "system", "content": prompt},
                    {"role": "user", "content": bug_report}
                ]
            )
            parsed = response.choices[0].message['content'].strip()
            return {"status": "success", "output": {"parsed_report": parsed}}
        except Exception as e:
            logger.error(f"OpenAI API error: {e}")
            return {"status": "error", "message": str(e), "output": None}

    def validate_bug_report(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        model = node_data.get('model', 'gpt-3.5-turbo')
        prompt = node_data.get('prompt', 'Validate the following bug report:')
        parsed_report = node_data.get('parsed_report', '')
        try:
            response = openai.ChatCompletion.create(
                model=model,
                messages=[
                    {"role": "system", "content": prompt},
                    {"role": "user", "content": parsed_report}
                ]
            )
            validation = response.choices[0].message['content'].strip()
            return {"status": "success", "output": {"is_complete": validation == "Complete"}}
        except Exception as e:
            logger.error(f"OpenAI API error: {e}")
            return {"status": "error", "message": str(e), "output": None}



import configparser
import os
import re
import json
from typing import Dict, Any, List, Union
from pathlib import Path

class ActfileParserError(Exception):
    """Custom exception for Actfile parsing errors."""
    pass

class ActfileParser:
    def __init__(self, file_path: Union[str, Path]):
        self.file_path = Path(file_path)
        self.parsed_data: Dict[str, Any] = {}

    def parse(self) -> Dict[str, Any]:
        if not self.file_path.exists():
            raise ActfileParserError(f"Actfile not found: {self.file_path}")

        try:
            with open(self.file_path, 'r') as file:
                content = file.read()
            
            sections = self._split_sections(content)
            
            self.parsed_data = {
                "workflow": self._parse_section(sections.get('workflow', '')),
                "nodes": self._parse_nodes(sections),
                "edges": self._parse_edges(sections.get('edges', '')),
                "dependencies": self._parse_dependencies(sections.get('dependencies', '')),
                "env": self._parse_env(sections.get('env', '')),
                "settings": self._parse_settings(sections.get('settings', ''))
            }

            self._validate_parsed_data()

        except Exception as e:
            raise ActfileParserError(f"Error parsing Actfile: {e}")

        return self.parsed_data

    def _split_sections(self, content: str) -> Dict[str, str]:
        pattern = r'\[([^\]]+)\](.*?)(?=\n\[|$)'
        matches = re.findall(pattern, content, re.DOTALL)
        return {match[0].strip(): match[1].strip() for match in matches}

    def _parse_section(self, content: str) -> Dict[str, Any]:
        lines = content.split('\n')
        section_data = {}
        for line in lines:
            if '=' in line:
                key, value = line.split('=', 1)
                section_data[key.strip()] = self._parse_value(value.strip())
        return section_data

    def _parse_value(self, value: str) -> Any:
        if value.startswith('{') and value.endswith('}'):
            try:
                return json.loads(value)
            except json.JSONDecodeError:
                pass
        elif value.startswith('"""') and value.endswith('"""'):
            return value.strip('"""').strip()
        return value

    def _parse_nodes(self, sections: Dict[str, str]) -> Dict[str, Dict[str, Any]]:
        nodes = {}
        for section_name, content in sections.items():
            if section_name.startswith('node:'):
                node_name = section_name.split('node:')[1]
                node_data = self._parse_section(content)
                if 'type' not in node_data:
                    raise ActfileParserError(f"Node '{node_name}' must have a 'type' field")
                nodes[node_name] = node_data
        return nodes

    def _parse_edges(self, content: str) -> Dict[str, List[str]]:
        edges = {}
        for line in content.split('\n'):
            line = line.strip()
            if line and '=' in line:
                parts = line.split('=', 1)
                if len(parts) == 2:
                    source, targets = parts
                    source = source.strip()
                    if source.startswith(';'):
                        raise ActfileParserError(f"Invalid edge definition: '{line}'. Remove the leading semicolon.")
                    targets = [t.strip() for t in targets.split(',') if t.strip()]
                    edges[source] = targets
                else:
                    raise ActfileParserError(f"Invalid edge definition: '{line}'. Expected format: 'source = target1, target2'")
        return edges
    def _parse_dependencies(self, content: str) -> Dict[str, List[str]]:
        dependencies = {}
        for line in content.split('\n'):
            if '=' in line:
                node_type, deps = line.split('=', 1)
                dependencies[node_type.strip()] = [d.strip() for d in deps.split(',')]
        return dependencies

    def _parse_env(self, content: str) -> Dict[str, str]:
        env_vars = {}
        for line in content.split('\n'):
            if '=' in line:
                key, value = line.split('=', 1)
                key = key.strip()
                value = value.strip()
                if value.startswith('${') and value.endswith('}'):
                    env_var_name = value[2:-1]
                    env_vars[key] = os.environ.get(env_var_name, '')
                else:
                    env_vars[key] = value
        return env_vars

    def _parse_settings(self, content: str) -> Dict[str, Any]:
        settings = {}
        for line in content.split('\n'):
            if '=' in line:
                key, value = line.split('=', 1)
                key = key.strip()
                value = value.strip()
                try:
                    settings[key] = json.loads(value)
                except json.JSONDecodeError:
                    settings[key] = value
        return settings

    def _validate_parsed_data(self):
        # Validate that all nodes mentioned in edges exist in the nodes section
        all_nodes = set(self.parsed_data['nodes'].keys())
        for source, targets in self.parsed_data['edges'].items():
            if source not in all_nodes:
                raise ActfileParserError(f"Edge source node '{source}' not defined in nodes")
            for target in targets:
                if target not in all_nodes:
                    raise ActfileParserError(f"Edge target node '{target}' not defined in nodes")

        # Validate that all dependencies refer to existing node types
        node_types = set(node['type'] for node in self.parsed_data['nodes'].values())
        for dep_node_type in self.parsed_data['dependencies'].keys():
            if dep_node_type not in node_types:
                raise ActfileParserError(f"Dependency defined for non-existent node type: {dep_node_type}")


    def get_node_dependencies(self, node_type: str) -> List[str]:
        return self.parsed_data['dependencies'].get(node_type, [])

    def get_workflow_name(self) -> str:
        return self.parsed_data['workflow']['name']

    def get_workflow_description(self) -> str:
        return self.parsed_data['workflow']['description']

    def get_start_node(self) -> str:
        # Use the start_node specified in the workflow section
        return self.parsed_data['workflow'].get('start_node')

    def get_node_successors(self, node_name: str) -> List[str]:
        return self.parsed_data['edges'].get(node_name, [])

    def get_env_var(self, key: str, default: Any = None) -> Any:
        return self.parsed_data['env'].get(key, default)

    def get_setting(self, key: str, default: Any = None) -> Any:
        return self.parsed_data['settings'].get(key, default)

    def to_json(self) -> str:
        return json.dumps(self.parsed_data, indent=2)

    def validate_node_config(self, node_name: str, required_fields: List[str]) -> None:
        node_config = self.parsed_data['nodes'].get(node_name)
        if not node_config:
            raise ActfileParserError(f"Node '{node_name}' not found in configuration")
        
        for field in required_fields:
            if field not in node_config:
                raise ActfileParserError(f"Missing required field '{field}' in node '{node_name}' configuration")

    @staticmethod
    def find_actfile(start_dir: Union[str, Path] = None) -> Path:
        """Find the Actfile by searching in the current directory and its parents."""
        current_dir = Path(start_dir or os.getcwd())
        while True:
            actfile_path = current_dir / 'Actfile'
            if actfile_path.is_file():
                return actfile_path
            if current_dir.parent == current_dir:
                raise ActfileParserError("Actfile not found in current directory or any parent directory")
            current_dir = current_dir.parent



import logging
from typing import Dict, Any

logger = logging.getLogger(__name__)

class GenericNode:
    def execute(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        logger.info("Executing GenericNode for operation")
        operation = node_data.get('operation', '').lower()
        input_data = node_data.get('input', {})

        if operation == 'get':
            return self.handle_get(input_data)
        elif operation == 'post':
            return self.handle_post(input_data)
        elif operation == 'put':
            return self.handle_put(input_data)
        elif operation == 'delete':
            return self.handle_delete(input_data)
        else:
            return {"status": "error", "message": f"Unsupported operation: {operation}", "output": None}

    def handle_get(self, input_data):
        return {"status": "success", "message": "GET operation handled", "output": input_data}

    def handle_post(self, input_data):
        return {"status": "success", "message": "POST operation handled", "output": input_data}

    def handle_put(self, input_data):
        return {"status": "success", "message": "PUT operation handled", "output": input_data}

    def handle_delete(self, input_data):
        return {"status": "success", "message": "DELETE operation handled", "output": input_data}



import os
import json
import logging
from typing import Dict, Any, List
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError
import re

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

class SlackNode:
    def __init__(self):
        logger.info("Initializing SlackNode")
        self.client = None

    def extract_text(self, input_text: str) -> str:
        try:
            parsed = json.loads(input_text)
            if isinstance(parsed, dict):
                return str(parsed.get('value', input_text))
            elif isinstance(parsed, str):
                return parsed
        except json.JSONDecodeError:
            pass
        return input_text

    def resolve_path_placeholders(self, text: str, node_data: Dict[str, Any]) -> str:
        pattern = re.compile(r"\{\{(.*?)\}\}")
        matches = pattern.findall(text)
        
        for match in matches:
            parts = match.split('.')
            node_id = parts[0]
            path = '.'.join(parts[1:])
            value = self.fetch_value(node_id, path, node_data)
            if value is not None:
                text = text.replace(f"{{{{{match}}}}}", str(value))
        
        return text

    def fetch_value(self, node_id: str, path: str, node_data: Dict[str, Any]) -> Any:
        try:
            node_result = node_data.get('input', {}).get('result', {})
            for part in path.split('.'):
                node_result = node_result.get(part, None)
                if node_result is None:
                    break
            return node_result
        except Exception as e:
            logger.error(f"Failed to fetch value for {node_id}.{path}: {str(e)}")
            return None

    def execute(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        logger.info("Starting execution of SlackNode")
        logger.info(f"Received node_data: {json.dumps(self.log_safe_node_data(node_data), indent=2)}")
        
        slack_token = node_data.get('slack_token')
        method = node_data.get('method')
        params = node_data.get('params', {})

        if not slack_token:
            logger.error("Missing Slack token")
            return {"status": "error", "message": "Missing Slack token"}

        if not method:
            logger.error("Missing method")
            return {"status": "error", "message": "Missing method"}

        try:
            self.client = WebClient(token=slack_token)
            
            # Resolve any placeholders in the params
            resolved_params = {k: self.resolve_path_placeholders(str(v), node_data) for k, v in params.items()}
            
            # Call the appropriate method
            method_to_call = getattr(self.client, method, None)
            if method_to_call:
                response = method_to_call(**resolved_params)
                result = self.serialize_slack_response(response)
            else:
                logger.error(f"Unknown method: {method}")
                return {"status": "error", "message": f"Unknown method: {method}"}

            return {
                "status": "success",
                "result": result
            }

        except SlackApiError as e:
            error_msg = f"Slack API error: {e.response['error']}"
            logger.error(error_msg)
            return {"status": "error", "message": error_msg}
        except Exception as e:
            error_msg = f"Unexpected error during execution: {str(e)}"
            logger.error(error_msg)
            return {"status": "error", "message": error_msg}

    def serialize_slack_response(self, response) -> Dict[str, Any]:
        """
        Serialize the SlackResponse object into a dictionary.
        """
        return {
            "ok": response.get("ok", False),
            "channel": response.get("channel"),
            "ts": response.get("ts"),
            "message": response.get("message", {})
        }
    
    @staticmethod
    def log_safe_node_data(node_data):
        if isinstance(node_data, dict):
            safe_data = {k: ('[REDACTED]' if k == 'slack_token' else v) for k, v in node_data.items()}
        else:
            safe_data = node_data
        return safe_data



    # Web API Methods
    def admin_analytics_getFile(self, **kwargs):
        return self.client.admin_analytics_getFile(**kwargs)

    def admin_apps_activities_list(self, **kwargs):
        return self.client.admin_apps_activities_list(**kwargs)

    def admin_apps_approve(self, **kwargs):
        return self.client.admin_apps_approve(**kwargs)

    def admin_apps_approved_list(self, **kwargs):
        return self.client.admin_apps_approved_list(**kwargs)

    def admin_apps_clearResolution(self, **kwargs):
        return self.client.admin_apps_clearResolution(**kwargs)

    def admin_apps_config_lookup(self, **kwargs):
        return self.client.admin_apps_config_lookup(**kwargs)

    def admin_apps_config_set(self, **kwargs):
        return self.client.admin_apps_config_set(**kwargs)

    def admin_apps_requests_list(self, **kwargs):
        return self.client.admin_apps_requests_list(**kwargs)

    def admin_apps_requests_cancel(self, **kwargs):
        return self.client.admin_apps_requests_cancel(**kwargs)

    def admin_apps_restrict(self, **kwargs):
        return self.client.admin_apps_restrict(**kwargs)

    def admin_apps_restricted_list(self, **kwargs):
        return self.client.admin_apps_restricted_list(**kwargs)

    def admin_apps_uninstall(self, **kwargs):
        return self.client.admin_apps_uninstall(**kwargs)

    def admin_audit_anomaly_allow_list(self, **kwargs):
        return self.client.admin_audit_anomaly_allow_list(**kwargs)

    def admin_auth_policy_assignEntities(self, **kwargs):
        return self.client.admin_auth_policy_assignEntities(**kwargs)

    def admin_auth_policy_getEntities(self, **kwargs):
        return self.client.admin_auth_policy_getEntities(**kwargs)

    def admin_auth_policy_removeEntities(self, **kwargs):
        return self.client.admin_auth_policy_removeEntities(**kwargs)

    def admin_barriers_create(self, **kwargs):
        return self.client.admin_barriers_create(**kwargs)

    def admin_barriers_delete(self, **kwargs):
        return self.client.admin_barriers_delete(**kwargs)

    def admin_barriers_list(self, **kwargs):
        return self.client.admin_barriers_list(**kwargs)

    def admin_barriers_update(self, **kwargs):
        return self.client.admin_barriers_update(**kwargs)

    def admin_conversations_archive(self, **kwargs):
        return self.client.admin_conversations_archive(**kwargs)

    def admin_conversations_convertToPrivate(self, **kwargs):
        return self.client.admin_conversations_convertToPrivate(**kwargs)

    def admin_conversations_create(self, **kwargs):
        return self.client.admin_conversations_create(**kwargs)

    def admin_conversations_delete(self, **kwargs):
        return self.client.admin_conversations_delete(**kwargs)

    def admin_conversations_disconnectShared(self, **kwargs):
        return self.client.admin_conversations_disconnectShared(**kwargs)

    def admin_conversations_ekm_listOriginalConnectedChannelInfo(self, **kwargs):
        return self.client.admin_conversations_ekm_listOriginalConnectedChannelInfo(**kwargs)

    def admin_conversations_getConversationPrefs(self, **kwargs):
        return self.client.admin_conversations_getConversationPrefs(**kwargs)

    def admin_conversations_getTeams(self, **kwargs):
        return self.client.admin_conversations_getTeams(**kwargs)

    def admin_conversations_invite(self, **kwargs):
        return self.client.admin_conversations_invite(**kwargs)

    def admin_conversations_rename(self, **kwargs):
        return self.client.admin_conversations_rename(**kwargs)

    def admin_conversations_restrictAccess_addGroup(self, **kwargs):
        return self.client.admin_conversations_restrictAccess_addGroup(**kwargs)

    def admin_conversations_restrictAccess_listGroups(self, **kwargs):
        return self.client.admin_conversations_restrictAccess_listGroups(**kwargs)

    def admin_conversations_restrictAccess_removeGroup(self, **kwargs):
        return self.client.admin_conversations_restrictAccess_removeGroup(**kwargs)

    def admin_conversations_search(self, **kwargs):
        return self.client.admin_conversations_search(**kwargs)

    def admin_conversations_setConversationPrefs(self, **kwargs):
        return self.client.admin_conversations_setConversationPrefs(**kwargs)

    def admin_conversations_setTeams(self, **kwargs):
        return self.client.admin_conversations_setTeams(**kwargs)

    def admin_conversations_unarchive(self, **kwargs):
        return self.client.admin_conversations_unarchive(**kwargs)

    def admin_emoji_add(self, **kwargs):
        return self.client.admin_emoji_add(**kwargs)

    def admin_emoji_addAlias(self, **kwargs):
        return self.client.admin_emoji_addAlias(**kwargs)

    def admin_emoji_list(self, **kwargs):
        return self.client.admin_emoji_list(**kwargs)

    def admin_emoji_remove(self, **kwargs):
        return self.client.admin_emoji_remove(**kwargs)

    def admin_emoji_rename(self, **kwargs):
        return self.client.admin_emoji_rename(**kwargs)

    def admin_functions_list(self, **kwargs):
        return self.client.admin_functions_list(**kwargs)

    def admin_functions_permissions_lookup(self, **kwargs):
        return self.client.admin_functions_permissions_lookup(**kwargs)

    def admin_functions_permissions_set(self, **kwargs):
        return self.client.admin_functions_permissions_set(**kwargs)

    def admin_inviteRequests_approve(self, **kwargs):
        return self.client.admin_inviteRequests_approve(**kwargs)

    def admin_inviteRequests_approved_list(self, **kwargs):
        return self.client.admin_inviteRequests_approved_list(**kwargs)

    def admin_inviteRequests_denied_list(self, **kwargs):
        return self.client.admin_inviteRequests_denied_list(**kwargs)

    def admin_inviteRequests_deny(self, **kwargs):
        return self.client.admin_inviteRequests_deny(**kwargs)

    def admin_inviteRequests_list(self, **kwargs):
        return self.client.admin_inviteRequests_list(**kwargs)

    def admin_roles_addAssignments(self, **kwargs):
        return self.client.admin_roles_addAssignments(**kwargs)

    def admin_roles_listAssignments(self, **kwargs):
        return self.client.admin_roles_listAssignments(**kwargs)

    def admin_roles_removeAssignments(self, **kwargs):
        return self.client.admin_roles_removeAssignments(**kwargs)

    def admin_teams_admins_list(self, **kwargs):
        return self.client.admin_teams_admins_list(**kwargs)

    def admin_teams_create(self, **kwargs):
        return self.client.admin_teams_create(**kwargs)

    def admin_teams_list(self, **kwargs):
        return self.client.admin_teams_list(**kwargs)

    def admin_teams_owners_list(self, **kwargs):
        return self.client.admin_teams_owners_list(**kwargs)

    def admin_teams_settings_info(self, **kwargs):
        return self.client.admin_teams_settings_info(**kwargs)

    def admin_teams_settings_setDefaultChannels(self, **kwargs):
        return self.client.admin_teams_settings_setDefaultChannels(**kwargs)

    def admin_teams_settings_setDescription(self, **kwargs):
        return self.client.admin_teams_settings_setDescription(**kwargs)

    def admin_teams_settings_setDiscoverability(self, **kwargs):
        return self.client.admin_teams_settings_setDiscoverability(**kwargs)

    def admin_teams_settings_setIcon(self, **kwargs):
        return self.client.admin_teams_settings_setIcon(**kwargs)

    def admin_teams_settings_setName(self, **kwargs):
        return self.client.admin_teams_settings_setName(**kwargs)

    def admin_usergroups_addChannels(self, **kwargs):
        return self.client.admin_usergroups_addChannels(**kwargs)

    def admin_usergroups_addTeams(self, **kwargs):
        return self.client.admin_usergroups_addTeams(**kwargs)

    def admin_usergroups_listChannels(self, **kwargs):
        return self.client.admin_usergroups_listChannels(**kwargs)

    def admin_usergroups_removeChannels(self, **kwargs):
        return self.client.admin_usergroups_removeChannels(**kwargs)

    def admin_users_assign(self, **kwargs):
        return self.client.admin_users_assign(**kwargs)

    def admin_users_invite(self, **kwargs):
        return self.client.admin_users_invite(**kwargs)

    def admin_users_list(self, **kwargs):
        return self.client.admin_users_list(**kwargs)

    def admin_users_remove(self, **kwargs):
        return self.client.admin_users_remove(**kwargs)

    def admin_users_setAdmin(self, **kwargs):
        return self.client.admin_users_setAdmin(**kwargs)

    def admin_users_setExpiration(self, **kwargs):
        return self.client.admin_users_setExpiration(**kwargs)

    def admin_users_setOwner(self, **kwargs):
        return self.client.admin_users_setOwner(**kwargs)

    def admin_users_setRegular(self, **kwargs):
        return self.client.admin_users_setRegular(**kwargs)

    def admin_users_session_invalidate(self, **kwargs):
        return self.client.admin_users_session_invalidate(**kwargs)

    def admin_users_session_list(self, **kwargs):
        return self.client.admin_users_session_list(**kwargs)

    def admin_users_session_reset(self, **kwargs):
        return self.client.admin_users_session_reset(**kwargs)

    def admin_users_session_resetBulk(self, **kwargs):
        return self.client.admin_users_session_resetBulk(**kwargs)

    def admin_users_unsupportedVersions_export(self, **kwargs):
        return self.client.admin_users_unsupportedVersions_export(**kwargs)

    def admin_workflows_collaborators_add(self, **kwargs):
        return self.client.admin_workflows_collaborators_add(**kwargs)

    def admin_workflows_collaborators_remove(self, **kwargs):
        return self.client.admin_workflows_collaborators_remove(**kwargs)

    def admin_workflows_permissions_lookup(self, **kwargs):
        return self.client.admin_workflows_permissions_lookup(**kwargs)

    def admin_workflows_search(self, **kwargs):
        return self.client.admin_workflows_search(**kwargs)

    def admin_workflows_triggers_types_permissions_list(self, **kwargs):
        return self.client.admin_workflows_triggers_types_permissions_list(**kwargs)

    def admin_workflows_unpublish(self, **kwargs):
        return self.client.admin_workflows_unpublish(**kwargs)

    def api_test(self, **kwargs):
        return self.client.api_test(**kwargs)

    def apps_activities_list(self, **kwargs):
        return self.client.apps_activities_list(**kwargs)

    def apps_auth_external_create(self, **kwargs):
        return self.client.apps_auth_external_create(**kwargs)

    def apps_auth_external_delete(self, **kwargs):
        return self.client.apps_auth_external_delete(**kwargs)

    def apps_connections_open(self, **kwargs):
        return self.client.apps_connections_open(**kwargs)

    def apps_datastore_delete(self, **kwargs):
        return self.client.apps_datastore_delete(**kwargs)

    def apps_datastore_get(self, **kwargs):
        return self.client.apps_datastore_get(**kwargs)

    def apps_datastore_put(self, **kwargs):
        return self.client.apps_datastore_put(**kwargs)

    def apps_datastore_query(self, **kwargs):
        return self.client.apps_datastore_query(**kwargs)

    def apps_datastore_update(self, **kwargs):
        return self.client.apps_datastore_update(**kwargs)

    def apps_event_authorizations_list(self, **kwargs):
        return self.client.apps_event_authorizations_list(**kwargs)

    def apps_manifest_create(self, **kwargs):
        return self.client.apps_manifest_create(**kwargs)

    def apps_manifest_delete(self, **kwargs):
        return self.client.apps_manifest_delete(**kwargs)

    def apps_manifest_export(self, **kwargs):
        return self.client.apps_manifest_export(**kwargs)

    def apps_manifest_update(self, **kwargs):
        return self.client.apps_manifest_update(**kwargs)

    def apps_manifest_validate(self, **kwargs):
        return self.client.apps_manifest_validate(**kwargs)

    def apps_uninstall(self, **kwargs):
        return self.client.apps_uninstall(**kwargs)

    def assistant_threads_ask(self, **kwargs):
        return self.client.assistant_threads_ask(**kwargs)

    def auth_revoke(self, **kwargs):
        return self.client.auth_revoke(**kwargs)

    def auth_test(self, **kwargs):
        return self.client.auth_test(**kwargs)

    def auth_teams_list(self, **kwargs):
        return self.client.auth_teams_list(**kwargs)

    def bookmarks_add(self, **kwargs):
        return self.client.bookmarks_add(**kwargs)

    def bookmarks_edit(self, **kwargs):
        return self.client.bookmarks_edit(**kwargs)

    def bookmarks_list(self, **kwargs):
        return self.client.bookmarks_list(**kwargs)

    def bookmarks_remove(self, **kwargs):
        return self.client.bookmarks_remove(**kwargs)

    def bots_info(self, **kwargs):
        return self.client.bots_info(**kwargs)

    def calls_add(self, **kwargs):
        return self.client.calls_add(**kwargs)

    def calls_end(self, **kwargs):
        return self.client.calls_end(**kwargs)

    def calls_info(self, **kwargs):
        return self.client.calls_info(**kwargs)

    def calls_participants_add(self, **kwargs):
        return self.client.calls_participants_add(**kwargs)

    def calls_participants_remove(self, **kwargs):
        return self.client.calls_participants_remove(**kwargs)

    def calls_update(self, **kwargs):
        return self.client.calls_update(**kwargs)

    def canvases_access_add(self, **kwargs):
        return self.client.canvases_access_add(**kwargs)

    def canvases_access_remove(self, **kwargs):
        return self.client.canvases_access_remove(**kwargs)

    def canvases_create(self, **kwargs):
        return self.client.canvases_create(**kwargs)

    def canvases_duplicate(self, **kwargs):
        return self.client.canvases_duplicate(**kwargs)

    def canvases_get(self, **kwargs):
        return self.client.canvases_get(**kwargs)

    def canvases_list(self, **kwargs):
        return self.client.canvases_list(**kwargs)

    def canvases_rename(self, **kwargs):
        return self.client.canvases_rename(**kwargs)

    def canvases_sections_add(self, **kwargs):
        return self.client.canvases_sections_add(**kwargs)

    def canvases_sections_get(self, **kwargs):
        return self.client.canvases_sections_get(**kwargs)

    def canvases_sections_list(self, **kwargs):
        return self.client.canvases_sections_list(**kwargs)

    def canvases_sections_remove(self, **kwargs):
        return self.client.canvases_sections_remove(**kwargs)

    def canvases_sections_rename(self, **kwargs):
        return self.client.canvases_sections_rename(**kwargs)

    def channels_archive(self, **kwargs):
        return self.client.channels_archive(**kwargs)

    def channels_create(self, **kwargs):
        return self.client.channels_create(**kwargs)

    def channels_history(self, **kwargs):
        return self.client.channels_history(**kwargs)

    def channels_info(self, **kwargs):
        return self.client.channels_info(**kwargs)

    def channels_invite(self, **kwargs):
        return self.client.channels_invite(**kwargs)

    def channels_join(self, **kwargs):
        return self.client.channels_join(**kwargs)

    def channels_kick(self, **kwargs):
        return self.client.channels_kick(**kwargs)

    def channels_leave(self, **kwargs):
        return self.client.channels_leave(**kwargs)

    def channels_list(self, **kwargs):
        return self.client.channels_list(**kwargs)

    def channels_mark(self, **kwargs):
        return self.client.channels_mark(**kwargs)

    def channels_rename(self, **kwargs):
        return self.client.channels_rename(**kwargs)

    def channels_replies(self, **kwargs):
        return self.client.channels_replies(**kwargs)

    def channels_setPurpose(self, **kwargs):
        return self.client.channels_setPurpose(**kwargs)

    def channels_setTopic(self, **kwargs):
        return self.client.channels_setTopic(**kwargs)

    def channels_unarchive(self, **kwargs):
        return self.client.channels_unarchive(**kwargs)

    def chat_delete(self, **kwargs):
        return self.client.chat_delete(**kwargs)

    def chat_deleteScheduledMessage(self, **kwargs):
        return self.client.chat_deleteScheduledMessage(**kwargs)

    def chat_getPermalink(self, **kwargs):
        return self.client.chat_getPermalink(**kwargs)

    def chat_meMessage(self, **kwargs):
        return self.client.chat_meMessage(**kwargs)

    def chat_postEphemeral(self, **kwargs):
        return self.client.chat_postEphemeral(**kwargs)

    def chat_postMessage(self, **kwargs):
        return self.client.chat_postMessage(**kwargs)

    def chat_scheduleMessage(self, **kwargs):
        return self.client.chat_scheduleMessage(**kwargs)

    def chat_scheduledMessages_list(self, **kwargs):
        return self.client.chat_scheduledMessages_list(**kwargs)

    def chat_unfurl(self, **kwargs):
        return self.client.chat_unfurl(**kwargs)

    def chat_update(self, **kwargs):
        return self.client.chat_update(**kwargs)

    def conversations_archive(self, **kwargs):
        return self.client.conversations_archive(**kwargs)

    def conversations_close(self, **kwargs):
        return self.client.conversations_close(**kwargs)

    def conversations_create(self, **kwargs):
        return self.client.conversations_create(**kwargs)

    def conversations_history(self, **kwargs):
        return self.client.conversations_history(**kwargs)

    def conversations_info(self, **kwargs):
        return self.client.conversations_info(**kwargs)

    def conversations_invite(self, **kwargs):
        return self.client.conversations_invite(**kwargs)

    def conversations_join(self, **kwargs):
        return self.client.conversations_join(**kwargs)

    def conversations_kick(self, **kwargs):
        return self.client.conversations_kick(**kwargs)

    def conversations_leave(self, **kwargs):
        return self.client.conversations_leave(**kwargs)

    def conversations_list(self, **kwargs):
        return self.client.conversations_list(**kwargs)

    def conversations_mark(self, **kwargs):
        return self.client.conversations_mark(**kwargs)

    def conversations_members(self, **kwargs):
        return self.client.conversations_members(**kwargs)

    def conversations_open(self, **kwargs):
        return self.client.conversations_open(**kwargs)

    def conversations_rename(self, **kwargs):
        return self.client.conversations_rename(**kwargs)

    def conversations_replies(self, **kwargs):
        return self.client.conversations_replies(**kwargs)

    def conversations_setPurpose(self, **kwargs):
        return self.client.conversations_setPurpose(**kwargs)

    def conversations_setTopic(self, **kwargs):
        return self.client.conversations_setTopic(**kwargs)

    def conversations_unarchive(self, **kwargs):
        return self.client.conversations_unarchive(**kwargs)

    def dialog_open(self, **kwargs):
        return self.client.dialog_open(**kwargs)

    def dnd_endDnd(self, **kwargs):
        return self.client.dnd_endDnd(**kwargs)

    def dnd_endSnooze(self, **kwargs):
        return self.client.dnd_endSnooze(**kwargs)

    def dnd_info(self, **kwargs):
        return self.client.dnd_info(**kwargs)

    def dnd_setSnooze(self, **kwargs):
        return self.client.dnd_setSnooze(**kwargs)

    def dnd_teamInfo(self, **kwargs):
        return self.client.dnd_teamInfo(**kwargs)

    def emoji_list(self, **kwargs):
        return self.client.emoji_list(**kwargs)

    def files_comments_delete(self, **kwargs):
        return self.client.files_comments_delete(**kwargs)

    def files_delete(self, **kwargs):
        return self.client.files_delete(**kwargs)

    def files_info(self, **kwargs):
        return self.client.files_info(**kwargs)

    def files_list(self, **kwargs):
        return self.client.files_list(**kwargs)

    def files_remote_add(self, **kwargs):
        return self.client.files_remote_add(**kwargs)

    def files_remote_info(self, **kwargs):
        return self.client.files_remote_info(**kwargs)

    def files_remote_list(self, **kwargs):
        return self.client.files_remote_list(**kwargs)

    def files_remote_remove(self, **kwargs):
        return self.client.files_remote_remove(**kwargs)

    def files_remote_share(self, **kwargs):
        return self.client.files_remote_share(**kwargs)

    def files_remote_update(self, **kwargs):
        return self.client.files_remote_update(**kwargs)

    def files_revokePublicURL(self, **kwargs):
        return self.client.files_revokePublicURL(**kwargs)

    def files_sharedPublicURL(self, **kwargs):
        return self.client.files_sharedPublicURL(**kwargs)

    def files_upload(self, **kwargs):
        return self.client.files_upload(**kwargs)

    def functions_workflows_steps_list(self, **kwargs):
        return self.client.functions_workflows_steps_list(**kwargs)

    def functions_workflows_steps_update(self, **kwargs):
        return self.client.functions_workflows_steps_update(**kwargs)

    def groups_archive(self, **kwargs):
        return self.client.groups_archive(**kwargs)

    def groups_create(self, **kwargs):
        return self.client.groups_create(**kwargs)

    def groups_createChild(self, **kwargs):
        return self.client.groups_createChild(**kwargs)

    def groups_history(self, **kwargs):
        return self.client.groups_history(**kwargs)

    def groups_info(self, **kwargs):
        return self.client.groups_info(**kwargs)

    def groups_invite(self, **kwargs):
        return self.client.groups_invite(**kwargs)

    def groups_kick(self, **kwargs):
        return self.client.groups_kick(**kwargs)

    def groups_leave(self, **kwargs):
        return self.client.groups_leave(**kwargs)

    def groups_list(self, **kwargs):
        return self.client.groups_list(**kwargs)

    def groups_mark(self, **kwargs):
        return self.client.groups_mark(**kwargs)

    def groups_open(self, **kwargs):
        return self.client.groups_open(**kwargs)

    def groups_rename(self, **kwargs):
        return self.client.groups_rename(**kwargs)

    def groups_replies(self, **kwargs):
        return self.client.groups_replies(**kwargs)

    def groups_setPurpose(self, **kwargs):
        return self.client.groups_setPurpose(**kwargs)

    def groups_setTopic(self, **kwargs):
        return self.client.groups_setTopic(**kwargs)

    def groups_unarchive(self, **kwargs):
        return self.client.groups_unarchive(**kwargs)

    def im_close(self, **kwargs):
        return self.client.im_close(**kwargs)

    def im_history(self, **kwargs):
        return self.client.im_history(**kwargs)

    def im_list(self, **kwargs):
        return self.client.im_list(**kwargs)

    def im_mark(self, **kwargs):
        return self.client.im_mark(**kwargs)

    def im_open(self, **kwargs):
        return self.client.im_open(**kwargs)

    def im_replies(self, **kwargs):
        return self.client.im_replies(**kwargs)

    def migration_exchange(self, **kwargs):
        return self.client.migration_exchange(**kwargs)

    def mpim_close(self, **kwargs):
        return self.client.mpim_close(**kwargs)

    def mpim_history(self, **kwargs):
        return self.client.mpim_history(**kwargs)

    def mpim_list(self, **kwargs):
        return self.client.mpim_list(**kwargs)

    def mpim_mark(self, **kwargs):
        return self.client.mpim_mark(**kwargs)

    def mpim_open(self, **kwargs):
        return self.client.mpim_open(**kwargs)

    def mpim_replies(self, **kwargs):
        return self.client.mpim_replies(**kwargs)

    def oauth_access(self, **kwargs):
        return self.client.oauth_access(**kwargs)

    def oauth_v2_access(self, **kwargs):
        return self.client.oauth_v2_access(**kwargs)

    def openid_connect_token(self, **kwargs):
        return self.client.openid_connect_token(**kwargs)

    def openid_connect_userInfo(self, **kwargs):
        return self.client.openid_connect_userInfo(**kwargs)

    def pins_add(self, **kwargs):
        return self.client.pins_add(**kwargs)

    def pins_list(self, **kwargs):
        return self.client.pins_list(**kwargs)

    def pins_remove(self, **kwargs):
        return self.client.pins_remove(**kwargs)

    def reactions_add(self, **kwargs):
        return self.client.reactions_add(**kwargs)

    def reactions_get(self, **kwargs):
        return self.client.reactions_get(**kwargs)

    def reactions_list(self, **kwargs):
        return self.client.reactions_list(**kwargs)

    def reactions_remove(self, **kwargs):
        return self.client.reactions_remove(**kwargs)

    def reminders_add(self, **kwargs):
        return self.client.reminders_add(**kwargs)

    def reminders_complete(self, **kwargs):
        return self.client.reminders_complete(**kwargs)

    def reminders_delete(self, **kwargs):
        return self.client.reminders_delete(**kwargs)

    def reminders_info(self, **kwargs):
        return self.client.reminders_info(**kwargs)

    def reminders_list(self, **kwargs):
        return self.client.reminders_list(**kwargs)

    def rtm_connect(self, **kwargs):
        return self.client.rtm_connect(**kwargs)

    def rtm_start(self, **kwargs):
        return self.client.rtm_start(**kwargs)

    def search_all(self, **kwargs):
        return self.client.search_all(**kwargs)

    def search_files(self, **kwargs):
        return self.client.search_files(**kwargs)

    def search_messages(self, **kwargs):
        return self.client.search_messages(**kwargs)

    def stars_add(self, **kwargs):
        return self.client.stars_add(**kwargs)

    def stars_list(self, **kwargs):
        return self.client.stars_list(**kwargs)

    def stars_remove(self, **kwargs):
        return self.client.stars_remove(**kwargs)

    def team_accessLogs(self, **kwargs):
        return self.client.team_accessLogs(**kwargs)

    def team_billableInfo(self, **kwargs):
        return self.client.team_billableInfo(**kwargs)

    def team_info(self, **kwargs):
        return self.client.team_info(**kwargs)

    def team_integrationLogs(self, **kwargs):
        return self.client.team_integrationLogs(**kwargs)

    def team_profile_get(self, **kwargs):
        return self.client.team_profile_get(**kwargs)

    def tooling_tokens_rotate(self, **kwargs):
        return self.client.tooling_tokens_rotate(**kwargs)

    def usergroups_create(self, **kwargs):
        return self.client.usergroups_create(**kwargs)

    def usergroups_disable(self, **kwargs):
        return self.client.usergroups_disable(**kwargs)

    def usergroups_enable(self, **kwargs):
        return self.client.usergroups_enable(**kwargs)

    def usergroups_list(self, **kwargs):
        return self.client.usergroups_list(**kwargs)

    def usergroups_update(self, **kwargs):
        return self.client.usergroups_update(**kwargs)

    def usergroups_users_list(self, **kwargs):
        return self.client.usergroups_users_list(**kwargs)

    def usergroups_users_update(self, **kwargs):
        return self.client.usergroups_users_update(**kwargs)

    def users_conversations(self, **kwargs):
        return self.client.users_conversations(**kwargs)

    def users_deletePhoto(self, **kwargs):
        return self.client.users_deletePhoto(**kwargs)

    def users_getPresence(self, **kwargs):
        return self.client.users_getPresence(**kwargs)

    def users_identity(self, **kwargs):
        return self.client.users_identity(**kwargs)

    def users_info(self, **kwargs):
        return self.client.users_info(**kwargs)

    def users_list(self, **kwargs):
        return self.client.users_list(**kwargs)

    def users_lookupByEmail(self, **kwargs):
        return self.client.users_lookupByEmail(**kwargs)

    def users_profile_get(self, **kwargs):
        return self.client.users_profile_get(**kwargs)

    def users_profile_set(self, **kwargs):
        return self.client.users_profile_set(**kwargs)

    def users_setActive(self, **kwargs):
        return self.client.users_setActive(**kwargs)

    def users_setPhoto(self, **kwargs):
        return self.client.users_setPhoto(**kwargs)

    def users_setPresence(self, **kwargs):
        return self.client.users_setPresence(**kwargs)

    def views_open(self, **kwargs):
        return self.client.views_open(**kwargs)

    def views_publish(self, **kwargs):
        return self.client.views_publish(**kwargs)

    def views_push(self, **kwargs):
        return self.client.views_push(**kwargs)

    def views_update(self, **kwargs):
        return self.client.views_update(**kwargs)

    def workflows_stepCompleted(self, **kwargs):
        return self.client.workflows_stepCompleted(**kwargs)

    def workflows_stepFailed(self, **kwargs):
        return self.client.workflows_stepFailed(**kwargs)

    def workflows_updateStep(self, **kwargs):
        return self.client.workflows_updateStep(**kwargs)


SlackNodeNode = SlackNode

if __name__ == "__main__":
    test_data = {
        "slack_token": os.environ.get("SLACK_TOKEN"),
        "method": "chat_postMessage",
        "params": {
            "channel": "#general",
            "text": "Hello from SlackNode!"
        }
    }

    node = SlackNode()
    result = node.execute(test_data)
    print(json.dumps(result, indent=2))


# from .slack_node import SlackNode
# from .openai_node import OpenAINode
# from .gitlab_node import GitLabNode
# from .ai_node import AINode
# from .new_node import NewNode
from .chatmodels_node import ChatModelsNode


import logging
from typing import Dict, Any
import gitlab
import os

logger = logging.getLogger(__name__)

class GitLabNode:
    def __init__(self):
        logger.info("Initializing GitLabNode")
        gitlab_token = os.environ.get('GITLAB_API_TOKEN')
        gitlab_project_id = os.environ.get('GITLAB_PROJECT_ID')
        if not gitlab_token:
            raise ValueError("GITLAB_API_TOKEN environment variable not set")
        if not gitlab_project_id:
            raise ValueError("GITLAB_PROJECT_ID environment variable not set")
        self.gl = gitlab.Gitlab('https://gitlab.com', private_token=gitlab_token)
        self.project = self.gl.projects.get(gitlab_project_id)

    def execute(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        operation = node_data.get('operation')
        if operation == 'createIssue':
            return self.create_issue(node_data)
        elif operation == 'assignIssue':
            return self.assign_issue(node_data)
        elif operation == 'addLabel':
            return self.add_label(node_data)
        elif operation == 'monitorIssue':
            return self.monitor_issue(node_data)
        elif operation == 'closeIssue':
            return self.close_issue(node_data)
        elif operation == 'reopenIssue':
            return self.reopen_issue(node_data)
        else:
            raise ValueError(f"Unsupported operation: {operation}")

    def create_issue(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        title = node_data.get('title', 'New Issue')
        description = node_data.get('description', '')
        try:
            issue = self.project.issues.create({'title': title, 'description': description})
            return {"status": "success", "output": {"issue_id": issue.id}}
        except Exception as e:
            logger.error(f"GitLab API error: {e}")
            return {"status": "error", "message": str(e), "output": None}

    def assign_issue(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        issue_id = node_data.get('issue_id')
        assignee_id = node_data.get('assignee_id')
        if not issue_id or not assignee_id:
            return {"status": "error", "message": "issue_id and assignee_id are required", "output": None}
        try:
            issue = self.project.issues.get(issue_id)
            issue.assignee_ids = [assignee_id]
            issue.save()
            return {"status": "success", "message": f"Issue {issue_id} assigned to user {assignee_id}"}
        except Exception as e:
            logger.error(f"GitLab API error: {e}")
            return {"status": "error", "message": str(e), "output": None}

    def add_label(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        issue_id = node_data.get('issue_id')
        labels = node_data.get('labels', [])
        if not issue_id:
            return {"status": "error", "message": "issue_id is required", "output": None}
        try:
            issue = self.project.issues.get(issue_id)
            issue.labels = labels
            issue.save()
            return {"status": "success", "message": f"Labels {labels} added to issue {issue_id}"}
        except Exception as e:
            logger.error(f"GitLab API error: {e}")
            return {"status": "error", "message": str(e), "output": None}

    def monitor_issue(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        issue_id = node_data.get('issue_id')
        if not issue_id:
            return {"status": "error", "message": "issue_id is required", "output": None}
        try:
            issue = self.project.issues.get(issue_id)
            status = issue.state
            return {"status": "success", "output": {"status": status}}
        except Exception as e:
            logger.error(f"GitLab API error: {e}")
            return {"status": "error", "message": str(e), "output": None}

    def close_issue(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        issue_id = node_data.get('issue_id')
        if not issue_id:
            return {"status": "error", "message": "issue_id is required", "output": None}
        try:
            issue = self.project.issues.get(issue_id)
            issue.state_event = 'close'
            issue.save()
            return {"status": "success", "message": f"Issue {issue_id} closed"}
        except Exception as e:
            logger.error(f"GitLab API error: {e}")
            return {"status": "error", "message": str(e), "output": None}

    def reopen_issue(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        issue_id = node_data.get('issue_id')
        if not issue_id:
            return {"status": "error", "message": "issue_id is required", "output": None}
        try:
            issue = self.project.issues.get(issue_id)
            issue.state_event = 'reopen'
            issue.save()
            return {"status": "success", "message": f"Issue {issue_id} reopened"}
        except Exception as e:
            logger.error(f"GitLab API error: {e}")
            return {"status": "error", "message": str(e), "output": None}



import os
import json
import logging
from typing import Dict, Any
from langchain_openai import ChatOpenAI
from langchain.schema import HumanMessage
import re

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logger = logging.getLogger(__name__)

class ChatModelsNode:
    def __init__(self):
        logger.info("Initializing ChatModelsNode")

    def extract_text(self, input_text: str) -> str:
        """
        Extract the actual text string from various possible input formats.
        """
        try:
            parsed = json.loads(input_text)
            if isinstance(parsed, dict):
                return str(parsed.get('value', input_text))
            elif isinstance(parsed, str):
                return parsed
        except json.JSONDecodeError:
            pass
        return input_text

    def resolve_path_placeholders(self, text: str, node_data: Dict[str, Any]) -> str:
        """
        Resolve any path placeholders in the input text before processing.
        """
        pattern = re.compile(r"\{\{(.*?)\}\}")
        matches = pattern.findall(text)
        
        for match in matches:
            parts = match.split('.')
            node_id = parts[0]
            path = '.'.join(parts[1:])
            value = self.fetch_value(node_id, path, node_data)
            if value is not None:
                text = text.replace(f"{{{{{match}}}}}", str(value))
        
        return text

    def fetch_value(self, node_id: str, path: str, node_data: Dict[str, Any]) -> Any:
        """
        Fetch the value from the node_data based on the node_id and path.
        """
        try:
            node_result = node_data.get('input', {}).get('result', {})
            for part in path.split('.'):
                node_result = node_result.get(part, None)
                if node_result is None:
                    break
            return node_result
        except Exception as e:
            logger.error(f"Failed to fetch value for {node_id}.{path}: {str(e)}")
            return None

    def execute(self, node_data: Dict[str, Any]) -> Dict[str, Any]:
        logger.info("Starting execution of ChatModelsNode")
        logger.info(f"Received node_data: {json.dumps(self.log_safe_node_data(node_data), indent=2)}")
        
        api_key = node_data.get('api_key')
        model_name = node_data.get('model', 'gpt-3.5-turbo')
        input_text = node_data.get('input_text')

        if not api_key:
            logger.error("Missing API key")
            return {"status": "error", "message": "Missing API key"}

        if not input_text:
            logger.error("Missing input text")
            return {"status": "error", "message": "Missing input text"}

        try:
            # Resolve any placeholders in the input text before processing
            resolved_input_text = self.resolve_path_placeholders(input_text, node_data)
            logger.info(f"Resolved input text: {resolved_input_text}")

            # Extract the actual text string
            actual_text = self.extract_text(resolved_input_text)
            logger.info(f"Extracted text: {actual_text}")

            self.model = ChatOpenAI(api_key=api_key, model_name=model_name)
            message = HumanMessage(content=actual_text)
            response = self.model.invoke([message])
            response_text = response.content
            logger.info(f"Model response: {response_text}")

            result = {
                "status": "success",
                "result": {
                    "response": response_text
                }
            }
            logger.info(f"Execution completed successfully. Result: {json.dumps(self.log_safe_node_data(result), indent=2)}")
            return result

        except Exception as e:
            error_msg = f"Unexpected error during execution: {str(e)}"
            logger.error(error_msg)
            return {"status": "error", "message": error_msg}

    @staticmethod
    def log_safe_node_data(node_data):
        if isinstance(node_data, dict):
            safe_data = {k: ('[REDACTED]' if k == 'api_key' else v) for k, v in node_data.items()}
        else:
            safe_data = node_data
        return safe_data

ChatModelsNodeNode = ChatModelsNode

if __name__ == "__main__":
    test_data = {
        "api_key": os.environ.get("OPENAI_API_KEY"),
        "input_text": "Explain the concept of quantum entanglement in simple terms.",
        "model": "gpt-4"
    }

    node = ChatModelsNode()
    result = node.execute(test_data)
    print(json.dumps(result, indent=2))


import os
import re
from setuptools import setup, find_packages

def get_and_update_version() -> str:
    init_file_path = os.path.join('act', '__init__.py')
    with open(init_file_path, 'r') as f:
        init_content = f.read()
    
    version_match = re.search(r"__version__\s*=\s*['\"]([^'\"]+)['\"]", init_content)
    if not version_match:
        raise RuntimeError("Unable to find version string.")
    
    current_version = version_match.group(1)
    
    # Split the version into its components
    version_parts = current_version.split('.')
    
    # Increment the patch version
    new_patch = int(version_parts[2]) + 1
    
    # Generate new version string
    new_version = f"{version_parts[0]}.{version_parts[1]}.{new_patch}"
    
    # Update the version in the file
    with open(init_file_path, 'w') as f:
        f.write(re.sub(r"__version__\s*=\s*['\"]([^'\"]+)['\"]",
                       f"__version__ = '{new_version}'", init_content))
    
    print(f"Updated version from {current_version} to {new_version}")
    return new_version

def get_long_description():
    with open('README.md', 'r', encoding='utf-8') as f:
        return f.read()

def clean_requires(requires):
    return [req for req in requires if not req.startswith('file://') and '@file://' not in req]

install_requires = [
    'aiohappyeyeballs==2.4.3',
    'aiohttp==3.10.10',
    'aiosignal==1.3.1',
    'annotated-types==0.7.0',
    'anyio==4.6.2.post1',
    'async-timeout==4.0.3',
    'attrs==24.2.0',
    'certifi==2024.8.30',
    'charset-normalizer==3.4.0',
    'dataclasses-json==0.6.7',
    'distro==1.9.0',
    'exceptiongroup==1.2.2',
    'frozenlist==1.4.1',
    'h11==0.14.0',
    'httpcore==1.0.6',
    'httpx==0.27.2',
    'idna==3.10',
    'jiter==0.6.1',
    'jsonpatch==1.33',
    'jsonpointer==3.0.0',
    'langchain==0.3.3',
    'langchain-community==0.3.2',
    'langchain-core==0.3.11',
    'langchain-openai==0.2.2',
    'langchain-text-splitters==0.3.0',
    'langsmith==0.1.135',
    'marshmallow==3.22.0',
    'multidict==6.1.0',
    'mypy-extensions==1.0.0',
    'numpy==1.26.4',
    'openai==1.51.2',
    'orjson==3.10.7',
    'packaging==24.1',
    'propcache==0.2.0',
    'pydantic==2.9.2',
    'pydantic-settings==2.6.0',
    'pydantic_core==2.23.4',
    'python-dotenv==1.0.1',
    'PyYAML==6.0.2',
    'regex==2024.9.11',
    'requests==2.32.3',
    'requests-toolbelt==1.0.0',
    'sniffio==1.3.1',
    'SQLAlchemy==2.0.36',
    'tenacity==8.5.0',
    'tiktoken==0.8.0',
    'tqdm==4.66.5',
    'typing-inspect==0.9.0',
    'typing_extensions==4.12.2',
    'urllib3==2.2.3',
    'yarl==1.15.4',
    'slack-sdk==3.33.1',

]

cleaned_requires = clean_requires(install_requires)
print(f"Original install_requires: {install_requires}")
print(f"Cleaned install_requires: {cleaned_requires}")

setup(
    name='act-workflow',
    version=get_and_update_version(),
    author='Taj Noah',
    author_email='mail@tajnoah.me',
    description='A library for executing workflow nodes based on Actfile configuration',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/tajalagawani/actpip',
    packages=find_packages(exclude=['tests', 'tests.*']),
    package_data={
        'act': ['py.typed', '*.pyi', '**/*.pyi'],
    },
    include_package_data=True,
    install_requires=cleaned_requires,
    extras_require={
        'dev': [
            'pytest>=6.2.5',
            'mypy>=0.910',
            'flake8>=3.9.2',
        ],
    },
    entry_points={
        'console_scripts': [
            'act=act.cli:main',
        ],
    },
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    python_requires='>=3.7',
    license='MIT',
)

print(f"Setup completed with version: {get_and_update_version()}")
